# -*- coding: utf-8 -*-
#
#  grapher.py
#  notebooks
#

import datetime as dt
import json
import random
import string
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Tuple

import pandas as pd
from dataclasses_json import LetterCase, dataclass_json
from dateutil.parser import parse

DATE_DISPLAY = {"yearIsDay": True, "zeroDay": "1970-01-01"}


class Chart:
    """
    Chart(df, title='Hello').mark_line().encode(x='dog', y='sheep')
    """

    def __init__(self, data: pd.DataFrame):
        self.data = data.copy()
        self.config = ChartConfig()
        self.x: Optional[str] = None
        self.y: Optional[str] = None
        self.entity: Optional[str] = None
        self.color: Optional[str] = None
        self.size: Optional[str] = None
        self.time_type = TimeType.YEAR
        self.selection: Optional[List[str]] = None
        self.timespan: Optional[Tuple[Any, Any]] = None
        self.x_unit: Optional[str] = None
        self.y_unit: Optional[str] = None

    def encode(
        self,
        x: Optional[str] = None,
        y: Optional[str] = None,
        entity: Optional[str] = None,
        color: Optional[str] = None,
        size: Optional[str] = None,
    ) -> "Chart":
        self.x = x
        self.y = y
        self.entity = entity
        self.color = color
        self.size = size

        # fail early if there's been a typo
        for col in [x, y, entity, color, size]:
            if col and col not in self.data.columns:
                raise ValueError(f"no such column: {col}")

        if x == "date":
            self.time_type = TimeType.DAY

        self.config.hide_legend = not entity

        return self

    def label(
        self, title: str = "", subtitle: str = "", source_desc: str = "", note: str = ""
    ) -> "Chart":
        self.config.title = title
        self.config.subtitle = subtitle
        self.config.source_desc = source_desc
        self.config.note = note
        return self

    def xaxis(
        self,
        label: Optional[str] = None,
        unit: Optional[str] = None,
        scale: Optional[Literal["linear", "log"]] = None,
        scale_control: Optional[bool] = None,
    ) -> "Chart":
        if label is not None:
            self.config.x_axis["label"] = label
        if unit is not None:
            self.x_unit = unit
        if scale is not None:
            self.config.x_axis["scaleType"] = scale
        if scale_control is not None:
            self.config.x_axis["canChangeScaleType"] = scale_control
        return self

    def yaxis(
        self,
        label: Optional[str] = None,
        unit: Optional[str] = None,
        scale: Optional[Literal["linear", "log"]] = None,
        scale_control: Optional[bool] = None,
    ) -> "Chart":
        if label is not None:
            self.config.y_axis["label"] = label
        if unit is not None:
            self.y_unit = unit
        if scale is not None:
            self.config.y_axis["scaleType"] = scale
        if scale_control is not None:
            self.config.y_axis["canChangeScaleType"] = scale_control
        return self

    def axis(
        self,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
        x_unit: Optional[str] = None,
        y_unit: Optional[str] = None,
        x_scale: Optional[Literal["linear", "log"]] = None,
        y_scale: Optional[Literal["linear", "log"]] = None,
        x_scale_control: Optional[bool] = None,
        y_scale_control: Optional[bool] = None,
    ) -> "Chart":
        """Configure both axes at once. For single-axis config, use xaxis() or yaxis()."""
        if x_label is not None:
            self.config.x_axis["label"] = x_label
        if y_label is not None:
            self.config.y_axis["label"] = y_label
        if x_unit is not None:
            self.x_unit = x_unit
        if y_unit is not None:
            self.y_unit = y_unit
        if x_scale is not None:
            self.config.x_axis["scaleType"] = x_scale
        if y_scale is not None:
            self.config.y_axis["scaleType"] = y_scale
        if x_scale_control is not None:
            self.config.x_axis["canChangeScaleType"] = x_scale_control
        if y_scale_control is not None:
            self.config.y_axis["canChangeScaleType"] = y_scale_control
        return self

    def mark_scatter(self) -> "Chart":
        self.config.type = "ScatterPlot"
        return self

    def mark_line(self) -> "Chart":
        self.config.type = "LineChart"
        return self

    def mark_bar(self, stacked=False) -> "Chart":
        if stacked:
            self.config.type = "StackedDiscreteBar"
        else:
            self.config.type = "DiscreteBar"
        return self

    def interact(
        self,
        allow_relative: Optional[bool] = None,
        scale_control: Optional[bool] = None,
        entity_control: Optional[bool] = None,
        enable_map: Optional[bool] = None,
    ) -> "Chart":
        if allow_relative is not None:
            self.config.hide_relative_toggle = False

        if scale_control is not None:
            # Update y_axis without overwriting existing settings
            self.config.y_axis["scaleType"] = "linear"
            self.config.y_axis["canChangeScaleType"] = scale_control

        if entity_control is not None:
            self.config.hide_entity_controls = not entity_control

        if enable_map:
            self.config.has_map_tab = True
            self.config.tab = "map"

        return self

    def select(
        self,
        entities: Optional[List[str]] = None,
        timespan: Optional[Tuple[Any, Any]] = None,
    ) -> "Chart":
        if entities:
            self.selection = entities

        if timespan:
            if isinstance(timespan, (str, int)):
                timespan = (timespan, None)
            self.timespan = timespan

        return self

    def transform(self, relative: bool) -> "Chart":
        self.config.stack_mode = "relative" if relative else "absolute"
        return self

    def filter(self, matching_entities_only: bool = True) -> "Chart":
        """Filter data to only show entities that have data for all dimensions."""
        self.config.matching_entities_only = matching_entities_only
        return self

    def _repr_html_(self):
        full_config = self.export()
        html = generate_iframe(full_config)
        return html

    def export(self, include_data=True) -> Dict[str, Any]:
        """Export the full config including data (for debugging)."""
        self.config.auto_improve()
        config = self.config.to_dict()  # type: ignore
        config.update(self.data_config().to_dict(self.config.type))
        config = prune(config)
        if not include_data:
            del config["owidDataset"]["data"]
        return config

    def data_config(self) -> "DataConfig":
        if not self.x or not self.y:
            raise ValueError("must provide an x and y encoding")
        return DataConfig.from_data(
            self.data,
            x=self.x,
            y=self.y,
            entity=self.entity,
            color=self.color,
            size=self.size,
            time_type=self.time_type,
            chart_type=self.config.type,
            selection=self.selection,
            timespan=self.timespan,
            x_unit=self.x_unit,
            y_unit=self.y_unit,
        )


class TimeType(Enum):
    DAY = "day"
    YEAR = "year"


ChartType = Literal["LineChart", "DiscreteBar", "ScatterPlot", "StackedDiscreteBar"]


@dataclass_json(letter_case=LetterCase.CAMEL)  # type: ignore
@dataclass
class ChartConfig:
    tab: str = "chart"
    title: str = ""
    subtitle: str = ""
    note: str = ""
    source_desc: str = ""
    hide_logo: bool = True
    is_published: bool = True
    type: ChartType = "LineChart"
    hide_title_annotation: bool = True
    hide_legend: bool = False
    hide_entity_controls: bool = True
    hide_relative_toggle: bool = True
    has_map_tab: bool = False
    stack_mode: Literal["relative", "absolute"] = "absolute"
    matching_entities_only: bool = False
    x_axis: dict = field(default_factory=dict)
    y_axis: dict = field(default_factory=dict)

    def auto_improve(self):
        if self.title and self.type == "LineChart":
            self.hide_title_annotation = False


@dataclass_json(letter_case=LetterCase.CAMEL)  # type: ignore
@dataclass
class Dimension:
    """
    "dimensions": [{"property": "y", "variableName": "population"}],
    """

    property: Literal["y", "x", "color"]
    variable_name: str
    display: Optional[dict] = field(default_factory=dict)


@dataclass
class DataConfig:
    """Stores data and column mappings for chart rendering."""

    df: pd.DataFrame
    x_col: str  # Time column (or x-axis for scatter)
    y_cols: List[str]  # Value column(s)
    entity_col: Optional[str]  # Entity/grouping column
    year_col: Optional[str]  # Year column (for scatter plots with time)
    color_col: Optional[str]  # Color column (for scatter plots, maps to colorSlug)
    size_col: Optional[str]  # Size column (for scatter plots)
    time_type: TimeType
    chart_type: ChartType
    selected_entity_names: List[str]
    min_time: Optional[Any] = None
    max_time: Optional[int] = None
    x_unit: Optional[str] = None  # Unit for x-axis variable
    y_unit: Optional[str] = None  # Unit for y-axis variable

    @classmethod
    def from_data(
        cls,
        df: pd.DataFrame,
        x: str,
        y: str,
        entity: Optional[str] = None,
        color: Optional[str] = None,
        size: Optional[str] = None,
        time_type: "TimeType" = TimeType.YEAR,
        chart_type: ChartType = "LineChart",
        selection: Optional[List[str]] = None,
        timespan: Optional[Tuple[Any, Any]] = None,
        x_unit: Optional[str] = None,
        y_unit: Optional[str] = None,
    ) -> "DataConfig":
        df = df.copy()

        year_col: Optional[str] = None
        color_col: Optional[str] = None

        # Determine entity column and selection
        if chart_type == "ScatterPlot":
            # For scatter: x and y are both value columns, entity is grouping
            entity_col = entity
            y_cols = [x, y]  # Both are "value" columns for scatter
            color_col = color  # Optional color encoding for scatter
            # Don't auto-select all entities for scatter plots - let grapher handle it
            if selection is None:
                selection = []
            # Check if dataframe has a year column
            if "year" in df.columns:
                year_col = "year"
        elif chart_type in ("DiscreteBar", "StackedDiscreteBar"):
            # For bar charts: y is entity, x is value
            entity_col = y
            y_cols = [x]
            if selection is None:
                selection = list(df[y].unique())
        else:
            # For line charts: x is time, y is value, entity is grouping
            entity_col = entity
            y_cols = [y]
            if selection is None:
                if entity:
                    selection = list(df[entity].unique())
                else:
                    selection = [y]  # Use column name as entity

        # Handle timespan
        min_time, max_time = None, None
        if timespan:
            if time_type == TimeType.DAY:
                timespan = _timespan_from_date(timespan)
            min_time, max_time = timespan

        # For scatter plots, always use 'latest'
        if chart_type == "ScatterPlot" and min_time is None:
            min_time = "latest"

        return DataConfig(
            df=df,
            x_col=x,
            y_cols=y_cols,
            entity_col=entity_col,
            year_col=year_col,
            color_col=color_col,
            size_col=size,
            time_type=time_type,
            chart_type=chart_type,
            selected_entity_names=selection,
            min_time=min_time,
            max_time=max_time,
            x_unit=x_unit,
            y_unit=y_unit,
        )

    def _get_dimensions(self) -> List[Dimension]:
        """Build dimension list based on chart type."""
        if self.chart_type == "ScatterPlot":
            # For scatter: y_cols[0] is x-axis, y_cols[1] is y-axis
            return [
                Dimension(property="y", variable_name=self.y_cols[1]),
                Dimension(property="x", variable_name=self.y_cols[0]),
            ]
        else:
            # For other charts: one dimension per y column
            return [Dimension(property="y", variable_name=col) for col in self.y_cols]

    def to_dict(self, chart_type: ChartType) -> Dict[Any, Any]:
        display = DATE_DISPLAY if self.time_type == TimeType.DAY else {}

        # Build metadata for columns
        metadata: Dict[str, Any] = {}
        for col in self.y_cols:
            col_display = display.copy()
            # For scatter plots: y_cols[0] is x-axis, y_cols[1] is y-axis
            if chart_type == "ScatterPlot":
                if col == self.y_cols[0] and self.x_unit:
                    col_display["unit"] = self.x_unit
                elif col == self.y_cols[1] and self.y_unit:
                    col_display["unit"] = self.y_unit
            else:
                # For other charts, apply y_unit to all y columns
                if self.y_unit:
                    col_display["unit"] = self.y_unit
            metadata[col] = {"display": col_display} if col_display else {"display": {}}

        # Rename entity column to entityName for OwidTable
        df = self.df.copy()
        if self.entity_col and self.entity_col in df.columns:
            df = df.rename(columns={self.entity_col: "entityName"})

        doc: Dict[str, Any] = {
            "selectedEntityNames": self.selected_entity_names,
            "owidDataset": {
                "data": df.to_dict(orient="list"),
                "metadata": metadata,
            },
            "dimensions": [d.to_dict() for d in self._get_dimensions()],  # type: ignore
            "chartTypes": [chart_type],
        }

        if self.min_time is not None:
            doc["minTime"] = self.min_time
        if self.max_time is not None:
            doc["maxTime"] = self.max_time
        if self.size_col is not None:
            doc["sizeSlug"] = self.size_col
        if self.color_col is not None:
            doc["colorSlug"] = self.color_col

        return doc


def _build_column_defs(config: Dict[str, Any]) -> str:
    """Build columnDefs array from metadata for OwidTable constructor."""
    owid_dataset = config.get("owidDataset", {})
    metadata = owid_dataset.get("metadata", {})

    column_defs = []
    for col_name, col_metadata in metadata.items():
        col_def = {"slug": col_name, "type": "Numeric"}

        # Extract display settings if present
        display = col_metadata.get("display", {})
        if display:
            col_def["display"] = display

        column_defs.append(col_def)

    # Return as JSON string for embedding in JavaScript
    return json.dumps(column_defs)


def _config_to_grapher(config: Dict[str, Any]) -> Dict[str, Any]:
    """Build GrapherState config from the export config."""
    dimensions = config.get("dimensions", [])
    chart_types = config.get("chartTypes", [])
    is_scatter = "ScatterPlot" in chart_types

    grapher_config: Dict[str, Any] = {
        "hideLogo": config.get("hideLogo", True),
        "selectedEntityNames": config.get("selectedEntityNames", []),
    }

    # Use chartTypes directly from config
    if config.get("chartTypes"):
        grapher_config["chartTypes"] = config["chartTypes"]

    # For scatter plots, separate x and y slugs
    if is_scatter:
        y_var_names = []
        x_var_name = None

        for dim in dimensions:
            var_name = dim.get("variableName")
            if var_name:
                if dim.get("property") == "y":
                    y_var_names.append(var_name)
                elif dim.get("property") == "x":
                    x_var_name = var_name

        if y_var_names:
            grapher_config["ySlugs"] = " ".join(y_var_names)
        if x_var_name:
            grapher_config["xSlug"] = x_var_name
    else:
        # For non-scatter plots, collect variable names from dimensions as ySlugs
        y_slugs = [
            dim.get("variableName") for dim in dimensions if dim.get("variableName")
        ]
        if y_slugs:
            grapher_config["ySlugs"] = " ".join(y_slugs)

    # Pass through common fields
    for field_name in [
        "title",
        "subtitle",
        "note",
        "sourceDesc",
        "hasMapTab",
        "tab",
        "stackMode",
        "minTime",
        "maxTime",
        "xAxis",
        "yAxis",
        "sizeSlug",
        "colorSlug",
    ]:
        if config.get(field_name):
            grapher_config[field_name] = config[field_name]

    # Pass through hide toggles and boolean flags
    for field_name in [
        "hideRelativeToggle",
        "hideEntityControls",
        "matchingEntitiesOnly",
    ]:
        if field_name in config:
            grapher_config[field_name] = config[field_name]

    return grapher_config


def generate_iframe(config: Dict[str, Any]) -> str:
    iframe_name = "".join(random.choice(string.ascii_lowercase) for _ in range(20))

    # Extract data for CSV and prepare config for GrapherState API
    csv_data = _config_to_csv(config)

    # Build column definitions from metadata
    column_defs = _build_column_defs(config)

    # Build grapher config from the config dict
    grapher_config = _config_to_grapher(config)

    iframe_contents = f"""
<!DOCTYPE html>
<html>
  <head>
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <link
      href="https://fonts.googleapis.com/css?family=Lato:300,400,400i,700,700i|Playfair+Display:400,700&display=swap"
      rel="stylesheet"
    />
    <link
      rel="stylesheet"
      href="https://expose-grapher-state.owid.pages.dev/assets/owid.css"
    />
    <style>
      body {{ margin: 0; padding: 0; }}
      figure {{ width: 100%; height: 100%; margin: 0; }}
      .error {{ color: red; padding: 20px; background: #fee; border-radius: 5px; }}
      /* Hide UI elements for cleaner notebook display */
      .grapher-share-button, .shareMenuOuter, .GrapherShareMenu,
      a[data-track-note="chart_click_explore"], .exploreDataButton,
      .originUrl, .SourcesFooter, .SourcesFooterHTML,
      .GrapherFooter__sourcesAndLicense {{ display: none !important; }}
    </style>
  </head>
  <body>
    <figure id="grapher-container"></figure>
    <script type="module" src="https://expose-grapher-state.owid.pages.dev/assets/owid.mjs"></script>
    <script type="module">
      await new Promise((resolve) => setTimeout(resolve, 500));

      const {{ Grapher, GrapherState, OwidTable, React, createRoot }} = window;
      const container = document.getElementById("grapher-container");

      if (!GrapherState || !OwidTable || !React || !createRoot) {{
        container.innerHTML = '<div class="error">Required exports not available. Check console.</div>';
        throw new Error("Required exports not available");
      }}

      const csvData = `{csv_data}`;
      const columnDefs = {column_defs};
      const table = new OwidTable(csvData, columnDefs);

      const grapherState = new GrapherState({{
        table: table,
        ...{json.dumps(grapher_config)},
        isConfigReady: true,
        isDataReady: true,
      }});

      const reactRoot = createRoot(container);
      reactRoot.render(React.createElement(Grapher, {{ grapherState }}));
    </script>
  </body>
</html>
"""  # noqa
    # Escape for the outer template literal (order matters: backslash first)
    iframe_contents = iframe_contents.replace("\\", "\\\\")
    iframe_contents = iframe_contents.replace("`", "\\`")
    iframe_contents = iframe_contents.replace("${", "\\${")
    iframe_contents = iframe_contents.replace("</script>", "<\\/script>")
    return f"""
        <div id="{iframe_name}_wrapper" style="position: relative; width: 100%; height: 600px;">
            <iframe id="{iframe_name}" style="width: 100%; height: 100%; border: 0px none; pointer-events: none;"></iframe>
        </div>
        <script>
            document.getElementById("{iframe_name}").contentDocument.write(`{iframe_contents}`);
            document.getElementById("{iframe_name}").contentDocument.close();
            // Enable interaction on click (wrapper captures click), disable on mouse leave
            document.getElementById("{iframe_name}_wrapper").addEventListener("click", function() {{
                document.getElementById("{iframe_name}").style.pointerEvents = "auto";
            }});
            document.getElementById("{iframe_name}").addEventListener("mouseleave", function() {{
                this.style.pointerEvents = "none";
            }});
        </script>
    """  # noqa


def _config_to_csv(config: Dict[str, Any]) -> str:
    """Convert the owidDataset format to CSV for OwidTable."""
    owid_dataset = config.get("owidDataset", {})
    data = owid_dataset.get("data", {})

    if not data:
        return ""

    # Convert dict of lists back to dataframe and then to CSV
    df = pd.DataFrame(data)
    return df.to_csv(index=False)


def prune(d: Dict[str, Any]) -> Dict[str, Any]:
    return {
        k: prune(v) if isinstance(v, dict) else v for k, v in d.items() if v is not None
    }


def _timespan_from_date(timespan: Tuple[str, str]) -> Tuple[int, int]:
    from_date_d = parse(timespan[0]).date()
    to_date_d = parse(timespan[1]).date()

    offset = dt.date(1970, 1, 1).toordinal()

    return (from_date_d.toordinal() - offset, to_date_d.toordinal() - offset)



# -*- coding: utf-8 -*-
#
#  notebook.py
#  owid-grapher-py
#

"""
Automatically generating notebooks from graphers.
"""

import json
from os import mkdir
from os.path import isdir, join
from typing import Iterator, List, Optional, Tuple

import click
import nbformat as nbf

# import jsonschema
import pandas as pd

from owid.site import get_owid_data, owid_data_to_frame

# all the types of charts we know how to translate back to python
WHITELIST_SCHEMA = {
    "$oneOf": [{"$ref": "/schemas/line_chart"}],
    "definitions": {
        "line_chart": {
            "$id": "/schemas/line_chart",
            "$allOf": [
                {
                    "type": "object",
                    "properties": {
                        "tab": {"enum": ["chart"]},
                    },
                },
                {"$ref": "/schemas/text_fields"},
            ],
        },
        "text_fields": {
            "$id": "/schemas/text_fields",
            "type": "object",
            "properties": {
                "title": {"type": "string"},
                "subtitle": {"type": "string"},
                "note": {"type": "string"},
                "sourceDesc": {"type": "string"},
            },
        },
    },
}


def translate_config(config: dict, data: pd.DataFrame) -> str:
    "Turn a grapher config into a python string describing the chart."
    # jsonschema.validate(config, WHITELIST_SCHEMA)

    chart_type = config.get("type", "LineChart")
    if chart_type == "LineChart":
        return translate_line_chart(config, data)

    raise UnsupportedChartType(chart_type)


def translate_line_chart(config: dict, data: pd.DataFrame) -> str:
    encoding = _gen_encoding(config, data)
    preselection, selection = _gen_selection(config, data)
    labels = _gen_labels(config)
    interaction = _gen_interaction(config)
    transform = _gen_transform(config)

    return f"""
grapher.Chart(
    data{preselection}
){encoding}{selection}{transform}{labels}{interaction}
""".strip()


def _gen_transform(config: dict) -> str:
    if config.get("stackMode") == "relative":
        return ".transform(\n     relative=True\n)"

    return ""


def _gen_encoding(config: dict, data: pd.DataFrame) -> str:
    if "date" in data:
        x = "date"
    else:
        x = "year"

    entity: Optional[str] = None
    if len(config["dimensions"]) > 1:
        entity = "variable"
    elif len(config.get("selectedData", [])) > 1:
        entity = "entity"
    elif len(config.get("selectedEntityNames", [])) > 1:
        entity = "entity"

    parts = [f'x="{x}"', 'y="value"']
    if entity:
        parts.append(f'entity="{entity}"')
    encoding = ",\n    ".join(parts)

    return f".encode(\n    {encoding}\n)"


def _gen_selection(config: dict, data: pd.DataFrame) -> Tuple[str, str]:
    """
    The config may select one variable and some of many entities, or it may select one entity and
    some of many variables.

    If we have multiple variables, pre-select the entity.
    """
    pre_selection, selection = _gen_entity_selection(config, data)

    min_time = config.get("minTime")
    max_time = config.get("maxTime")

    # don't set something that's automatic
    time = data["year"] if "year" in data.columns else data["date"]
    if min_time == time.min():
        min_time = None
    if max_time == time.max():
        max_time = None

    if pre_selection:
        if len(pre_selection) == 1:
            pre_selection_s = f'[data.entity == "{pre_selection[0]}"]'
        else:
            pre_selection_s = (
                ".query('entity in [\"" + '", "'.join(pre_selection) + "\"]')"
            )
    else:
        pre_selection_s = ""

    if selection and not min_time:
        middle = '",\n    "'.join(selection)
        selection_s = f""".select([
    "{middle}"
])"""
    elif min_time and not selection:
        selection_s = f""".select(
    timespan=({min_time}, {max_time})
)"""

    elif selection and min_time:
        middle = '",\n        "'.join(selection)
        selection_s = f""".select(
    entities=["{middle}"],
    timespan=({min_time}, {max_time})
)"""
    else:
        selection_s = ""

    return pre_selection_s, selection_s


def _gen_entity_selection(
    config: dict, data: pd.DataFrame
) -> Tuple[List[str], List[str]]:
    entities: List[str] = []

    if config.get("selectedEntityNames"):
        entities = config["selectedEntityNames"]

    elif config.get("selectedData") and len(config["selectedData"]) != len(
        data.entity.unique()
    ):
        selected_ids = [str(s["entityId"]) for s in config["selectedData"]]

        # requires an HTTP request
        owid_data = get_owid_data(config)

        entities = []
        for entity_id in selected_ids:
            try:
                entities.append(owid_data["entityKey"][entity_id]["name"])
            except KeyError:
                # some charts refer to entities that no longer exist
                # e.g. total-gov-expenditure-percapita-OECD
                continue
        entities = list(set(entities))

    # we have an actual selection
    if len(config["dimensions"]) > 1:
        # do entity pre-selection
        return entities, []

    return [], entities


def _gen_interaction(config: dict) -> str:
    parts = []

    entity_control = not config.get("hideEntityControls")
    if entity_control:
        parts.append("entity_control=True")

    scale_control = config.get("yAxis", {}).get("canChangeScaleType")
    if scale_control is not None:
        parts.append(f"scale_control={scale_control}")

    disable_relative = config.get("hideRelativeControls")
    if disable_relative is not None:
        parts.append(f"allow_relative={not disable_relative}")

    if config.get("hasMapTab"):
        parts.append("enable_map=True")

    if parts:
        return ".interact(\n    " + ",\n    ".join(parts) + "\n)"

    return ""


def _gen_labels(config: dict) -> str:
    to_snake = {"sourceDesc": "source_desc"}

    labels = {}
    for label in ["title", "subtitle", "sourceDesc", "note"]:
        if config.get(label):
            labels[to_snake.get(label, label)] = " ".join(config[label].split())

    if not labels:
        return ""

    return (
        ".label(\n    "
        + ",\n    ".join(f'{k}="{v}"' for k, v in labels.items())
        + "\n)"
    )


class UnsupportedChartType(Exception):
    pass


def generate_notebook(config: dict, path: str) -> None:
    "Generate a jupyter notebook for the given config in the provided folder."
    owid_data = get_owid_data(config)
    data = owid_data_to_frame(owid_data)
    py = translate_config(config, data)

    slug = config["slug"]
    title = config["title"]
    save_to_notebook(slug, title, py, path)


def save_to_notebook(slug: str, title: str, py: str, path: str) -> None:
    nb_file = join(path, f"{slug}.ipynb")

    nb = _new_notebook(slug, title, py)

    with open(nb_file, "w") as ostream:
        nbf.write(nb, ostream)


def _new_notebook(slug: str, title: str, py: str):
    nb = nbf.v4.new_notebook()
    nb["metadata"]["kernelspec"] = {
        "display_name": "Python 3 (ipykernel)",
        "language": "python",
        "name": "python3",
    }

    cells = []

    if title:
        cells.append(nbf.v4.new_markdown_cell(f"# {title}"))

    cells.append(
        nbf.v4.new_code_cell("import pandas as pd\nfrom owid import grapher, site")
    )

    cells.append(
        nbf.v4.new_code_cell(f'data = site.get_chart_data(slug="{slug}")\ndata.head()')
    )

    cells.append(nbf.v4.new_code_cell(py))

    nb["cells"] = cells

    return nb


@click.command()
@click.argument("input_file")
@click.argument("dest_path")
def main(input_file, dest_path):
    """
    Take a large list of configs in JSONL format, and attempt to render as many as we can to
    notebooks.
    """
    data_folder = join(dest_path, "_data")
    if not isdir(data_folder):
        mkdir(data_folder)

    i = 0
    for total, config in enumerate(iter_published(input_file), 1):
        slug = config["slug"]
        try:
            generate_notebook(config, dest_path)
            i += 1
            print(click.style(f"✓ [{i}/{total}] {slug}", fg="green"))
        except UnsupportedChartType:
            print(f"✗ [{i}/{total}] {slug}")

        except Exception as e:
            print(f"ERROR: {slug}")
            raise e

    print(f"Generated {i} notebooks successfully")


def iter_published(input_file: str) -> Iterator[dict]:
    for config in iter_jsonl(input_file):
        if not config.get("isPublished") or not config["slug"]:
            continue

        yield config


def iter_jsonl(input_file: str) -> Iterator[dict]:
    with open(input_file) as istream:
        for line in istream:
            yield json.loads(line)


if __name__ == "__main__":
    main()  # type: ignore



# -*- coding: utf-8 -*-
#
#  test_grapher.py
#  notebooks
#

import json

import pandas as pd

import owid.grapher as gr


def test_prune_empty():
    data = {}
    assert gr.prune(data) == data


def test_prune_single_layer():
    data = {"x": None, "y": 34}
    assert gr.prune(data) == {"y": 34}


def test_prune_recursive():
    data = {"x": None, "y": {"z": None, "f": 234, "g": {"k": None, "j": 123}}}
    expected = {"y": {"f": 234, "g": {"j": 123}}}
    assert gr.prune(data) == expected


def test_json_export():
    df = pd.DataFrame({"year": [2000, 2010, 2020], "population": [1234, 52342, 80123]})
    ch = (
        gr.Chart(df)
        .mark_line()
        .encode(x="year", y="population")
        .label("Too many Koalas?")
    )
    export = ch.export()
    assert export["title"] == "Too many Koalas?"
    assert export["type"] == "LineChart"
    assert export["selectedEntityNames"] == ["population"]

    # Check owidDataset structure - data is just the dataframe as dict
    data = export["owidDataset"]["data"]
    assert data["year"] == [2000, 2010, 2020]
    assert data["population"] == [1234, 52342, 80123]

    # Check metadata
    assert "population" in export["owidDataset"]["metadata"]


def test_scatter_plot_export():
    df = pd.DataFrame(
        {
            "gdp": [1000, 5000, 10000],
            "life_expectancy": [50, 70, 80],
            "country": ["Country A", "Country B", "Country C"],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_scatter()
        .encode(x="gdp", y="life_expectancy", entity="country")
        .label("Life expectancy vs. GDP")
    )
    config = ch.export()

    # Check key fields
    assert config["title"] == "Life expectancy vs. GDP"
    assert config["type"] == "ScatterPlot"
    assert config["selectedEntityNames"] == []  # Scatter plots don't auto-select
    assert config["minTime"] == "latest"

    # Check dimensions
    assert config["dimensions"] == [
        {"property": "y", "variableName": "life_expectancy", "display": {}},
        {"property": "x", "variableName": "gdp", "display": {}},
    ]

    # Check owidDataset structure - data is just the dataframe as dict
    data = config["owidDataset"]["data"]
    assert data["gdp"] == [1000, 5000, 10000]
    assert data["life_expectancy"] == [50, 70, 80]
    assert data["entityName"] == ["Country A", "Country B", "Country C"]

    # Check metadata
    assert "gdp" in config["owidDataset"]["metadata"]
    assert "life_expectancy" in config["owidDataset"]["metadata"]

    # Test CSV output
    from owid.grapher import _config_to_csv

    csv_output = _config_to_csv(config)
    # CSV should have all columns
    assert "gdp" in csv_output
    assert "life_expectancy" in csv_output
    assert "entityName" in csv_output
    assert "Country A" in csv_output


def test_scatter_plot_with_year():
    """Test scatter plot with year column - should preserve all rows."""
    df = pd.DataFrame(
        {
            "gdp_per_capita": [5000, 15000, 25000, 8000, 18000, 28000],
            "life_expectancy": [65, 72, 78, 68, 74, 79],
            "country": ["Australia"] * 3 + ["New Zealand"] * 3,
            "year": [2000, 2010, 2020, 2000, 2010, 2020],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_scatter()
        .encode(x="gdp_per_capita", y="life_expectancy", entity="country")
        .label(title="GDP vs Life Expectancy")
    )
    config = ch.export()

    # Check data preserves all 6 rows
    data = config["owidDataset"]["data"]
    assert len(data["gdp_per_capita"]) == 6
    assert len(data["life_expectancy"]) == 6
    assert len(data["entityName"]) == 6
    assert len(data["year"]) == 6

    # Check CSV output has all rows
    from owid.grapher import _config_to_csv

    csv_output = _config_to_csv(config)
    assert "year" in csv_output
    # Count rows (header + 6 data rows)
    assert csv_output.strip().count("\n") == 6


def test_scatter_plot_with_size():
    """Test scatter plot with size encoding."""
    df = pd.DataFrame(
        {
            "gdp": [1000, 5000, 10000],
            "life_expectancy": [50, 70, 80],
            "population": [10, 100, 50],
            "country": ["Country A", "Country B", "Country C"],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_scatter()
        .encode(x="gdp", y="life_expectancy", entity="country", size="population")
    )
    config = ch.export()

    # Check sizeSlug is set
    assert config["sizeSlug"] == "population"

    # Check data includes population column
    data = config["owidDataset"]["data"]
    assert data["population"] == [10, 100, 50]


def test_axis_labels():
    """Test axis label configuration."""
    df = pd.DataFrame(
        {
            "gdp": [1000, 5000, 10000],
            "life_expectancy": [50, 70, 80],
            "country": ["Country A", "Country B", "Country C"],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_scatter()
        .encode(x="gdp", y="life_expectancy", entity="country")
        .axis(x_label="GDP per capita ($)", y_label="Life expectancy (years)")
    )
    config = ch.export()

    assert config["xAxis"] == {"label": "GDP per capita ($)"}
    assert config["yAxis"] == {"label": "Life expectancy (years)"}


def test_axis_units():
    """Test axis unit configuration."""
    df = pd.DataFrame(
        {
            "gdp": [1000, 5000, 10000],
            "life_expectancy": [50, 70, 80],
            "country": ["Country A", "Country B", "Country C"],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_scatter()
        .encode(x="gdp", y="life_expectancy", entity="country")
        .axis(x_unit="$", y_unit="years")
    )
    config = ch.export()

    # Check that units are in metadata
    metadata = config["owidDataset"]["metadata"]
    assert metadata["gdp"]["display"]["unit"] == "$"
    assert metadata["life_expectancy"]["display"]["unit"] == "years"


def test_axis_units_line_chart():
    """Test axis unit configuration for line charts."""
    df = pd.DataFrame(
        {
            "year": [2000, 2010, 2020],
            "population": [1000, 2000, 3000],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_line()
        .encode(x="year", y="population")
        .axis(y_unit="millions")
    )
    config = ch.export()

    # Check that y_unit is in metadata for y column
    metadata = config["owidDataset"]["metadata"]
    assert metadata["population"]["display"]["unit"] == "millions"


def test_column_defs_generation():
    """Test that _build_column_defs correctly extracts units from metadata."""
    config = {
        "owidDataset": {
            "metadata": {
                "gdp": {"display": {"unit": "$"}},
                "life_expectancy": {"display": {"unit": "years"}},
            }
        }
    }

    column_defs_str = gr._build_column_defs(config)
    column_defs = json.loads(column_defs_str)

    # Check that we have two column definitions
    assert len(column_defs) == 2

    # Check structure
    gdp_def = next(d for d in column_defs if d["slug"] == "gdp")
    assert gdp_def["type"] == "Numeric"
    assert gdp_def["display"]["unit"] == "$"

    life_def = next(d for d in column_defs if d["slug"] == "life_expectancy")
    assert life_def["type"] == "Numeric"
    assert life_def["display"]["unit"] == "years"


def test_scatter_plot_iframe_with_units():
    """Test that scatter plot iframe includes columnDefs with units."""
    df = pd.DataFrame(
        {
            "gdp": [1000, 5000, 10000],
            "life_expectancy": [50, 70, 80],
            "country": ["Country A", "Country B", "Country C"],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_scatter()
        .encode(x="gdp", y="life_expectancy", entity="country")
        .axis(x_unit="$", y_unit="years")
    )

    # Get the iframe HTML
    html = ch._repr_html_()

    # Check that columnDefs is present in the HTML
    assert "const columnDefs" in html

    # Check that units are in the columnDefs
    assert '"unit": "$"' in html
    assert '"unit": "years"' in html

    # Check that OwidTable is called with columnDefs
    assert "new OwidTable(csvData, columnDefs)" in html


def test_axis_scale_configuration():
    """Test axis scale type and control configuration."""
    df = pd.DataFrame(
        {
            "gdp": [1000, 5000, 10000],
            "life_expectancy": [50, 70, 80],
            "country": ["Country A", "Country B", "Country C"],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_scatter()
        .encode(x="gdp", y="life_expectancy", entity="country")
        .axis(
            x_label="GDP per capita",
            x_scale="log",
            x_scale_control=True,
            y_label="Life expectancy",
        )
    )
    config = ch.export()

    # Check x-axis configuration
    assert config["xAxis"]["label"] == "GDP per capita"
    assert config["xAxis"]["scaleType"] == "log"
    assert config["xAxis"]["canChangeScaleType"] is True

    # Check y-axis configuration
    assert config["yAxis"]["label"] == "Life expectancy"


def test_axis_and_interact_compatibility():
    """Test that axis() and interact() work together without conflicts."""
    df = pd.DataFrame(
        {
            "year": [2000, 2010, 2020],
            "population": [1000, 2000, 3000],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_line()
        .encode(x="year", y="population")
        .axis(y_label="Population (millions)")
        .interact(scale_control=True)
    )
    config = ch.export()

    # Check that both label and scale control are present
    assert config["yAxis"]["label"] == "Population (millions)"
    assert config["yAxis"]["scaleType"] == "linear"
    assert config["yAxis"]["canChangeScaleType"] is True


def test_xaxis_yaxis_methods():
    """Test xaxis() and yaxis() methods provide cleaner API."""
    df = pd.DataFrame(
        {
            "gdp": [1000, 5000, 10000],
            "life_expectancy": [50, 70, 80],
            "country": ["Country A", "Country B", "Country C"],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_scatter()
        .encode(x="gdp", y="life_expectancy", entity="country")
        .xaxis(label="GDP per capita", unit="$", scale="log", scale_control=True)
        .yaxis(label="Life expectancy", unit="years")
    )
    config = ch.export()

    # Check x-axis configuration
    assert config["xAxis"]["label"] == "GDP per capita"
    assert config["xAxis"]["scaleType"] == "log"
    assert config["xAxis"]["canChangeScaleType"] is True

    # Check y-axis configuration
    assert config["yAxis"]["label"] == "Life expectancy"

    # Check units in metadata
    metadata = config["owidDataset"]["metadata"]
    assert metadata["gdp"]["display"]["unit"] == "$"
    assert metadata["life_expectancy"]["display"]["unit"] == "years"


def test_matching_entities_only():
    """Test matchingEntitiesOnly filter."""
    df = pd.DataFrame(
        {
            "gdp": [1000, 5000, 10000],
            "life_expectancy": [50, 70, 80],
            "country": ["Country A", "Country B", "Country C"],
        }
    )
    ch = (
        gr.Chart(df)
        .mark_scatter()
        .encode(x="gdp", y="life_expectancy", entity="country")
        .filter(matching_entities_only=True)
    )
    config = ch.export()

    # Check that matchingEntitiesOnly is set
    assert config["matchingEntitiesOnly"] is True



# -*- coding: utf-8 -*-
#
#  site.py
#  owid-grapher-py
#

"""
Tools for working with the live OWID grapher site.
"""

import datetime as dt
import json
from typing import Optional

import pandas as pd
import requests
from dateutil.parser import parse

DATA_URL = "https://api.ourworldindata.org/v1/indicators/{variable_id}.data.json"
METADATA_URL = (
    "https://api.ourworldindata.org/v1/indicators/{variable_id}.metadata.json"
)
GRAPHER_PREFIX = "https://ourworldindata.org/grapher/"
EPOCH_DATE = "2020-01-21"


def get_chart_config(url: str, force: bool = False) -> dict:
    "Get the internal OWID chart config for a chart URL."
    if not url.startswith(GRAPHER_PREFIX) and not force:
        raise Exception(f"not an OWID chart url: {url}")

    resp = requests.get(url)
    if resp.status_code != 200:
        raise Exception(f"got HTTP {resp.status_code} loading {url}")

    body = resp.content.decode("utf8")

    _, config, _ = body.split("//EMBEDDED_JSON")

    return json.loads(config)


def get_chart_data(
    url: Optional[str] = None, slug: Optional[str] = None
) -> pd.DataFrame:
    "Fetch the data from an OWID chart page as a data frame."
    if not url and not slug:
        raise ValueError("must provide an url or a slug")

    full_url: str = url or (GRAPHER_PREFIX + slug)  # type: ignore

    config = get_chart_config(full_url)
    owid_data = get_owid_data(config)
    return owid_data_to_frame(owid_data)


def owid_data_to_frame(owid_data: dict) -> pd.DataFrame:
    frames = []
    for variable in owid_data.values():
        # fetch metadata to get entity mapping
        meta = requests.get(METADATA_URL.format(variable_id=variable["id"])).json()
        entities = meta["dimensions"]["entities"]["values"]
        entity_map = {e["id"]: e["name"] for e in entities}

        df = pd.DataFrame(
            {
                "year": variable["years"],
                "entity": [entity_map[e] for e in variable["entities"]],
                "variable": meta["name"],
                "value": variable["values"],
            }
        )

        # handle date display
        if meta.get("display", {}).get("yearIsDay"):
            zero_day = parse(meta["display"].get("zeroDay", EPOCH_DATE)).date()
            df["date"] = df.pop("year").apply(lambda y: zero_day + dt.timedelta(days=y))
            df = df[["date", "entity", "variable", "value"]]

        frames.append(df)

    return pd.concat(frames)


def get_owid_data(config: dict) -> dict:
    variable_ids = [dim["variableId"] for dim in config["dimensions"]]
    owid_data = {}
    for variable_id in variable_ids:
        url = DATA_URL.format(variable_id=variable_id)
        data = requests.get(url).json()
        data["id"] = variable_id
        owid_data[variable_id] = data
    return owid_data
